@@ -1513,15 +1513,32 @@ static void perform_idle_server_maintenance(void)
     else if (idle_thread_count < min_spare_threads) {
         /* terminate the free list */
         if (free_length == 0) {
-            /* only report this condition once */
-            static int reported = 0;
+            /* No room for more children, might warn about configuration */
+            if (active_thread_count >= ap_daemons_limit * ap_threads_per_child) {
+                /* no threads are "inactive" - starting, stopping, etc. - which would confuse matters */
+                /* Are all threads in use?  Then we're really at MaxClients */
+                if (0 == idle_thread_count) {
+                    /* only report this condition once */
+                    static int reported = 0;
 
-            if (!reported) {
-                ap_log_error(APLOG_MARK, APLOG_ERR, 0,
-                             ap_server_conf,
-                             "server reached MaxClients setting, consider"
-                             " raising the MaxClients setting");
-                reported = 1;
+                    if (!reported) {
+                        ap_log_error(APLOG_MARK, APLOG_ERR, 0,
+                                     ap_server_conf,
+                                     "server reached MaxClients setting, consider"
+                                     " raising the MaxClients setting");
+                        reported = 1;
+                    }
+                } else {
+                    static int reported = 0;
+
+                    if (!reported) {
+                        ap_log_error(APLOG_MARK, APLOG_ERR, 0,
+                                     ap_server_conf,
+                                     "server is within MinSpareThreads of MaxClients, consider"
+                                     " raising the MaxClients setting");
+                        reported = 1;
+                    }
+                }
             }
             idle_spawn_rate = 1;
         }
@@ -1813,12 +1830,10 @@ int ap_mpm_run(apr_pool_t *_pconf, apr_pool_t *plog, server_rec *s)
 
             active_children = 0;
             for (index = 0; index < ap_daemons_limit; ++index) {
-                if (MPM_CHILD_PID(index) != 0) {
-                    if (kill(MPM_CHILD_PID(index), 0) == 0) {
-                            active_children = 1;
-                            /* Having just one child is enough to stay around */
-                            break;
-                    }
+                if (ap_mpm_safe_kill(MPM_CHILD_PID(index), 0) == APR_SUCCESS) {
+                    active_children = 1;
+                    /* Having just one child is enough to stay around */
+                    break;
                 }
             }
         } while (!shutdown_pending && active_children &&
